/* progress.c  -- progress bar (main routines)
   Copyright (C) 89, 90, 91, 1995-2002 Free Software Foundation.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Written by micha. */

/* 
 * Parts from:
 *
 * 1995 Timo Rinne <tri@iki.fi>, Tatu Ylonen <ylo@cs.hut.fi>
 *
 * As far as I am concerned, the code I have written for this software
 * can be used freely for any purpose.  Any derived versions of this
 * software must be clearly marked as such, and if the derived work is
 * incompatible with the protocol description in the RFC file, it must be
 * called by a name other than "ssh" or "Secure Shell".
 */
 
/*
 * Parts from:
 *
 * Copyright (c) 1999 Theo de Raadt.  All rights reserved.
 * Copyright (c) 1999 Aaron Campbell.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Parts from:
 *
 * Copyright (c) 1983, 1990, 1992, 1993, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "progress.h"
#include "du-core.h"
#include "dirname.h"

/* For progressmeter() -- number of seconds before xfer considered "stalled" */
#define STALLTIME       20
/* alarm() interval for updating progress meter */
#define PROGRESSTIME    1

#ifndef MAX
# define MAX(a,b) (((a)>(b))?(a):(b))
# define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef timersub
#define timersub(a, b, result)                                  \
   do {                                                         \
      (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;             \
      (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;          \
      if ((result)->tv_usec < 0) {                              \
         --(result)->tv_sec;                                    \
         (result)->tv_usec += 1000000;                          \
      }                                                         \
    } while (0)
#endif

/* Visual statistics about files as they are transferred. */
void progressmeter (int, int);

/* Returns width of the terminal (for progress meter calculations). */
int getttywidth (void);

/* Number of bytes of current file transferred so far. */
volatile off_t statbytes;

/* Total size in bytes to process. */
uintmax_t totalbytes = 0;

/* Name of current file being transferred. */
char *curfile;

/* Time a transfer started. */
static struct timeval start;

/* wrapper for signal interface */
typedef void (*mysig_t) (int);
mysig_t mysignal (int sig, mysig_t act);

/* Ensure all of data on socket comes through. f==read || f==write */
ssize_t atomicio (ssize_t (*)(), int, void *, size_t);

static void
updateprogressmeter (int ignore)
{
  int save_errno = errno;

  progressmeter (0,0);
  signal (SIGALRM, updateprogressmeter);
  alarm (PROGRESSTIME);
  errno = save_errno;
}

static int
foregroundproc (void)
{
  static pid_t pgrp = -1;
  int ctty_pgrp;

  if (pgrp == -1)
    pgrp = getpgrp ();

#ifdef HAVE_TCGETPGRP
  return ((ctty_pgrp = tcgetpgrp (STDOUT_FILENO)) != -1 && ctty_pgrp == pgrp);
#else
  return ((ioctl (STDOUT_FILENO, TIOCGPGRP, &ctty_pgrp) != -1 &&
	   ctty_pgrp == pgrp));
#endif
}

void
progressmeter (int flag, int interrupt)
{
  static const char prefixes[] = " KMGTP";
  static struct timeval lastupdate;
  static off_t lastsize;
  struct timeval now, td, wait;
  off_t cursize, abbrevsize;
  double elapsed;
  int ratio, barlength, i, remaining;
  char buf[512];

  if (flag == -1)
    {
      (void) gettimeofday (&start, (struct timezone *) 0);
      lastupdate = start;
      lastsize = 0;
    }

  if (foregroundproc () == 0)
    return;

  if (interrupt)
    {
      int len;
      
      len = (getttywidth()-1);
      snprintf (buf, sizeof (buf), "\r%*s\r", len, "");
      atomicio (write, fileno (stdout), buf, strlen (buf));
      return;
    }
  
  (void) gettimeofday (&now, (struct timezone *) 0);
  cursize = statbytes;

  if (totalbytes != 0)
    {
      ratio = 100.0 * cursize / totalbytes;
      ratio = MAX (ratio, 0);
      ratio = MIN (ratio, 100);
    }
  else
    ratio = 100;

  snprintf (buf, sizeof (buf), "\r%-20.20s %3d%% ", curfile, ratio);

  barlength = getttywidth () - 51;

  if (barlength > 0)
    {
      i = barlength * ratio / 100;
      snprintf (buf + strlen (buf), sizeof (buf) - strlen (buf),
		"|%.*s%*s|", i,
		"*******************************************************"
		"*******************************************************"
		"*******************************************************"
		"*******************************************************"
		"*******************************************************"
		"*******************************************************"
		"*******************************************************",
		barlength - i, "");
    }
  i = 0;
  abbrevsize = cursize;
  while (abbrevsize >= 100000 && i < sizeof (prefixes))
    {
      i++;
      abbrevsize >>= 10;
    }
  snprintf (buf + strlen (buf), sizeof (buf) - strlen (buf), " %5lu %c%c ",
	    (unsigned long) abbrevsize, prefixes[i],
	    prefixes[i] == ' ' ? ' ' : 'B');

  timersub (&now, &lastupdate, &wait);

  if (cursize > lastsize)
    {
      lastupdate = now;
      lastsize = cursize;
      if (wait.tv_sec >= STALLTIME)
	{
	  start.tv_sec += wait.tv_sec;
	  start.tv_usec += wait.tv_usec;
	}
      wait.tv_sec = 0;
    }

  timersub (&now, &start, &td);
  elapsed = td.tv_sec + (td.tv_usec / 1000000.0);

  if (flag != 1 && (statbytes <= 0 || elapsed <= 0.0 || cursize > totalbytes))
    {
      snprintf (buf + strlen (buf), sizeof (buf) - strlen (buf),
		"   --:-- ETA");
    }
  else if (wait.tv_sec >= STALLTIME)
    {
      snprintf (buf + strlen (buf), sizeof (buf) - strlen (buf),
		" - stalled -");
    }
  else
    {
      if (flag != 1)
	remaining = (int) (totalbytes / (statbytes / elapsed) - elapsed);
      else
	remaining = elapsed;

      i = remaining / 3600;
      if (i)
	snprintf (buf + strlen (buf), sizeof (buf) - strlen (buf), "%2d:", i);
      else
	snprintf (buf + strlen (buf), sizeof (buf) - strlen (buf), "   ");
      i = remaining % 3600;
      snprintf (buf + strlen (buf), sizeof (buf) - strlen (buf),
		"%02d:%02d%s", i / 60, i % 60, (flag != 1) ? " ETA" : "    ");
    }
  atomicio (write, fileno (stdout), buf, strlen (buf));

  if (flag == -1)
    {
      mysignal (SIGALRM, updateprogressmeter);
      alarm (PROGRESSTIME);
    }
  else if (flag == 1)
    {
      alarm (0);
      atomicio (write, fileno (stdout), "\n", 1);
      statbytes = 0;
    }
}

void
init_progressmeter (int n_files, char **file)
{
  char *args[n_files+1];
  int i;
  
  for (i = 0; i < n_files; i++)
    args[i] = file[i];
  
  args[n_files] = NULL;

  totalbytes = (du (args) * 512);
  statbytes = 0;
  progressmeter (-1,0);
}

void
update_progressmeter (char *file, off_t bytes)
{
  if (file)
    {
      (void)strip_trailing_slashes (file);
      curfile = base_name (file);
    }
    
  statbytes += bytes;
  progressmeter (0,0);
}

void 
interrupt_progressmeter (void)
{
  progressmeter (0,1);
}

void
close_progressmeter (void)
{
  if (statbytes < totalbytes && statbytes)
    update_progressmeter (NULL, totalbytes - statbytes);

  printf ("\n");
}

int
getttywidth (void)
{
  struct winsize winsize;

  if (ioctl (fileno (stdout), TIOCGWINSZ, &winsize) != -1)
    return (winsize.ws_col ? winsize.ws_col : 80);
  else
    return (80);
}

ssize_t
atomicio (ssize_t (*f)(), int fd, void *_s, size_t n)
{
  char *s = _s;
  ssize_t res, pos = 0;

  while (n > pos)
    {
      res = (f) (fd, s + pos, n - pos);
      switch (res)
	{
	case -1:
#ifdef EWOULDBLOCK
	  if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK)
#else
	  if (errno == EINTR || errno == EAGAIN)
#endif
	    continue;
	case 0:
	  return (res);
	default:
	  pos += res;
	}
    }
  return (pos);
}

mysig_t
mysignal (int sig, mysig_t act)
{
#ifdef HAVE_SIGACTION
  struct sigaction sa, osa;

  if (sigaction (sig, NULL, &osa) == -1)
    return (mysig_t) - 1;
  if (osa.sa_handler != act)
    {
      memset (&sa, 0, sizeof (sa));
      sigemptyset (&sa.sa_mask);
      sa.sa_flags = 0;
#if defined(SA_INTERRUPT)
      if (sig == SIGALRM)
	sa.sa_flags |= SA_INTERRUPT;
#endif
      sa.sa_handler = act;
      if (sigaction (sig, &sa, NULL) == -1)
	return (mysig_t) - 1;
    }
  return (osa.sa_handler);

#else
    return (signal(sig, act));
#endif
}
